#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import traceback
from typing import Set, Tuple

from th2_common.schema.message.message_listener import MessageListener
from th2_grpc_common.common_pb2 import MessageBatch

logger = logging.getLogger()


class SubscriptionManager(MessageListener):

    def __init__(self) -> None:
        self.listeners: Set[MessageListener] = set()

    def register(self, message_listener: MessageListener) -> None:
        self.listeners.add(message_listener)

    def unregister(self, message_listener: MessageListener) -> None:
        self.listeners.remove(message_listener)

    def handler(self, attributes: Tuple, message_batch: MessageBatch) -> None:
        for listener in self.listeners:
            try:
                listener.handler(attributes, message_batch)
            except Exception as e:
                logger.error(f'Cannot handle batch: {e} \n{"".join(traceback.format_tb(e.__traceback__))}')



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import traceback
from typing import Callable, Dict, List, Optional

from th2_act.util.cache import Cache
from th2_common.schema.message.message_listener import MessageListener
from th2_grpc_common.common_pb2 import Message, MessageBatch

logger = logging.getLogger()


StatusMessagesDict = Dict[int, List[Message]]


class ActReceiver:
    """Receives messages and put it in cache"""

    def __init__(self, prefilter: Optional[Callable], cache: Cache) -> None:
        if prefilter:
            self.prefilter = prefilter
            self.message_listener = ParsedMessageListener(self.process_incoming_messages_with_prefilter)
        else:
            self.message_listener = ParsedMessageListener(self.process_incoming_messages)

        self.cache: Cache = cache

    def process_incoming_messages(self, consumer_tag: str, message_batch: MessageBatch) -> None:
        try:
            logger.debug('Received MessageBatch with %i messages' % len(message_batch.messages))
            self.cache.add(message for message in message_batch.messages)

        except Exception as e:
            logger.error(f'Could not process incoming messages: {e}'
                         f'\n{"".join(traceback.format_tb(e.__traceback__))}')

    def process_incoming_messages_with_prefilter(self, consumer_tag: str, message_batch: MessageBatch) -> None:
        try:
            prefiltered_messages = [message for message in message_batch.messages if check(self.prefilter, message)]
            logger.debug('Received MessageBatch with %i messages, prefiltered %i messages'
                         % (len(message_batch.messages), len(prefiltered_messages)))
            self.cache.add(prefiltered_messages)

        except Exception as e:
            logger.error(f'Could not process incoming messages: {e}'
                         f'\n{"".join(traceback.format_tb(e.__traceback__))}')


class ParsedMessageListener(MessageListener):

    def __init__(self, handler_function: Callable) -> None:
        self.handler_function = handler_function

    def handler(self, consumer_tag: str, message_batch: MessageBatch) -> None:
        self.handler_function(consumer_tag, message_batch)


def check(condition: Callable[[Message], bool], message: Message) -> bool:
    try:
        return condition(message)
    except KeyError:
        return False



#   Copyright 2022-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from th2_act.util.subscription_manager import SubscriptionManager
from th2_common.schema.message.message_router import MessageRouter
from th2_grpc_check1.check1_service import Check1Service


class HandlerAttributes:
    """Contains parameters of ActHandler implementation. HandlerAttributes are used in RequestProcessor
    context manager. HandlerAttributes are created automatically when loading handlers.
    """

    __slots__ = ('check1_connector', 'subscription_manager', 'message_router', 'event_router')

    def __init__(self,
                 message_router: MessageRouter,
                 event_router: MessageRouter,
                 subscription_manager: SubscriptionManager,
                 check1_connector: Check1Service):
        self.message_router = message_router
        self.event_router = event_router
        self.check1_connector = check1_connector
        self.subscription_manager = subscription_manager



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from itertools import islice
from time import time
from typing import Callable, Dict, Iterable, Iterator, List, Optional, Tuple, Union

from th2_act import ActMessage
from th2_act.util.act_receiver import check
from th2_act.util.cache import Cache
from th2_act.util.grpc_context_manager import GRPCContextManager
from th2_grpc_common.common_pb2 import RequestStatus


class CacheFilter:
    MessageFiltersType = Union[Callable, Iterable[Callable]]

    def __init__(self, cache: Cache, grpc_context_manager: GRPCContextManager):
        self.cache = cache
        self.grpc_context_manager = grpc_context_manager
        self.last_filtered_message_index: Optional[int] = None

    def filter_single_message_immediately(
            self,
            check_previous_messages: bool,
            pass_on: Optional[MessageFiltersType] = None,
            fail_on: Optional[MessageFiltersType] = None) -> Tuple[Optional[int], ActMessage, float]:
        index, act_message = None, ActMessage()
        start_time = time()

        if self.grpc_context_manager.is_context_active():
            filtered_messages_slice = self.get_filtered_message_slice(
                message_filters=self._create_message_filters(pass_on=pass_on, fail_on=fail_on),
                messages_to_receive=1,
                start_index=self._get_start_index(check_previous_messages)
            )

            index, act_message = next(filtered_messages_slice, (index, act_message))

        runtime = time() - start_time
        self._update_last_filtered_message_index(index=index)

        return index, act_message, runtime

    def filter_single_message_within_timeout(
            self,
            timeout: Optional[float],
            check_previous_messages: bool,
            pass_on: Optional[MessageFiltersType] = None,
            fail_on: Optional[MessageFiltersType] = None) -> Tuple[Optional[int], ActMessage, float]:
        index, act_message = None, ActMessage()
        if timeout is None:
            timeout = self.grpc_context_manager.context_time_remaining()

        start_time = time()
        is_time_left = lambda: time() - start_time < timeout

        while index is None and is_time_left():
            if self.grpc_context_manager.is_context_active():
                filtered_messages_slice = self.get_filtered_message_slice(
                    message_filters=self._create_message_filters(pass_on=pass_on, fail_on=fail_on),
                    messages_to_receive=1,
                    start_index=self._get_start_index(check_previous_messages)
                )

                index, act_message = next(filtered_messages_slice, (index, act_message))
            else:
                break

        runtime = time() - start_time
        self._update_last_filtered_message_index(index=index)

        return index, act_message, runtime

    def filter_list_messages_immediately(
            self,
            check_previous_messages: bool,
            messages_to_receive: Optional[int] = None,
            pass_on: Optional[MessageFiltersType] = None,
            fail_on: Optional[MessageFiltersType] = None) -> Tuple[Optional[int], List[ActMessage], float]:
        index, act_messages = None, []
        start_time = time()

        if self.grpc_context_manager.is_context_active():
            filtered_messages_slice = self.get_filtered_message_slice(
                message_filters=self._create_message_filters(pass_on=pass_on, fail_on=fail_on),
                messages_to_receive=messages_to_receive,
                start_index=self._get_start_index(check_previous_messages)
            )

            for i, act_message in filtered_messages_slice:
                act_messages.append(act_message)
                index = i

        runtime = time() - start_time
        self._update_last_filtered_message_index(index=index)

        return index, act_messages, runtime

    def filter_list_messages_within_timeout(
            self,
            check_previous_messages: bool,
            messages_to_receive: int,
            timeout: Optional[float] = None,
            pass_on: Optional[MessageFiltersType] = None,
            fail_on: Optional[MessageFiltersType] = None) -> Tuple[Optional[int], List[ActMessage], float]:
        index, act_messages = None, []
        messages_to_filter = messages_to_receive
        if timeout is None:
            timeout = self.grpc_context_manager.context_time_remaining()

        start_time = time()
        is_time_left = lambda: time() - start_time < timeout

        while messages_to_filter > 0 and is_time_left():
            if self.grpc_context_manager.is_context_active():
                filtered_messages_slice = self.get_filtered_message_slice(
                    message_filters=self._create_message_filters(pass_on=pass_on, fail_on=fail_on),
                    messages_to_receive=messages_to_receive,
                    start_index=self._get_start_index(check_previous_messages)
                )

                for i, act_message in filtered_messages_slice:
                    act_messages.append(act_message)
                    messages_to_filter -= 1
                    index = i
            else:
                break

        runtime = time() - start_time
        self._update_last_filtered_message_index(index=index)

        return index, act_messages, runtime

    def get_filtered_message_slice(self,
                                   message_filters: Dict[Callable, int],
                                   messages_to_receive: Optional[int],
                                   start_index: int) -> Iterator[Tuple[int, ActMessage]]:
        return islice(
            (
                (index, ActMessage(message=message, status=status))
                for index, message in islice(enumerate(self.cache), start_index, self.cache.size)
                for condition, status in message_filters.items()
                if check(condition, message)
            ),
            messages_to_receive
        )

    def _create_message_filters(self,
                                pass_on: Optional[MessageFiltersType],
                                fail_on: Optional[MessageFiltersType]) -> Dict[Callable, int]:
        message_filters: Dict[Callable, int] = {}
        if isinstance(pass_on, Callable):  # type: ignore
            message_filters[pass_on] = RequestStatus.SUCCESS  # type: ignore
        elif isinstance(pass_on, Iterable):
            message_filters.update({message_filter: RequestStatus.SUCCESS for message_filter in pass_on})

        if isinstance(fail_on, Callable):  # type: ignore
            message_filters[fail_on] = RequestStatus.ERROR  # type: ignore
        elif isinstance(fail_on, Iterable):
            message_filters.update({message_filter: RequestStatus.ERROR for message_filter in fail_on})

        return message_filters

    def _get_start_index(self, check_previous_messages: bool) -> int:
        if check_previous_messages or self.last_filtered_message_index is None:
            return 0
        else:
            return self.last_filtered_message_index

    def _update_last_filtered_message_index(self, index: Optional[int]) -> None:
        if index is not None:
            self.last_filtered_message_index = index



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import traceback
from typing import Any, Callable, Iterable, List, Optional, Union

from th2_act.act_message import ActMessage, ActMessageUtil
from th2_act.grpc_method_attributes import GrpcMethodAttributes
from th2_act.handler_attributes import HandlerAttributes
from th2_act.util.act_receiver import ActReceiver
from th2_act.util.act_sender import ActSender
from th2_act.util.cache import Cache
from th2_act.util.cache_filter import CacheFilter
from th2_act.util.grpc_context_manager import GRPCContextManager
from th2_grpc_common.common_pb2 import Direction, EventID, Message, RequestStatus

logger = logging.getLogger()


class RequestProcessor:
    """A context manager used for sending requests and filtering messages from system under test.

    Mainly used when implementing gRPC methods in ActHandler class.

    Args:
        handler_attrs (HandlerAttributes): HandlerAttributes class instance. Generates by Act when loading handlers.
        method_attrs (GrpcMethodAttributes): GrpcMethodAttributes class instance. Initialize it with gRPC context,
            request event ID, gRPC method name and request description.
        prefilter (lambda or function, optional): Filter for messages that will be placed in cache.
            Defaults to None - cache will collect all messages.
    """

    MessageFiltersType = Union[Callable, Iterable[Callable]]

    def __init__(self,
                 handler_attrs: HandlerAttributes,
                 method_attrs: GrpcMethodAttributes,
                 prefilter: Optional[Callable] = None):
        self._subscription_manager = handler_attrs.subscription_manager
        self._grpc_context_manager = GRPCContextManager(context=method_attrs.context)

        self._cache: Cache = Cache()
        self._act_receiver: ActReceiver = ActReceiver(prefilter=prefilter, cache=self._cache)

        self._act_sender = ActSender(message_router=handler_attrs.message_router,
                                     event_router=handler_attrs.event_router,
                                     grpc_context_manager=self._grpc_context_manager,
                                     check1_connector=handler_attrs.check1_connector,
                                     method_attrs=method_attrs)
        self.checkpoint = self._act_sender.register_checkpoint()

        self.cache_filter = CacheFilter(cache=self._cache,
                                        grpc_context_manager=self._grpc_context_manager)

    def __enter__(self) -> Any:
        self._subscription_manager.register(message_listener=self._act_receiver.message_listener)
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> Optional[bool]:
        try:
            self._subscription_manager.unregister(message_listener=self._act_receiver.message_listener)
            self.clear_cache()
            if exc_type:
                logger.error(f'{exc_type} exception occurred in RequestProcessor: {exc_val}'
                             f'\n{"".join(traceback.format_tb(exc_tb))}')
                return True
            return None
        except Exception as e:
            logger.error(f'Could not stop subscriber monitor: {e}\n{"".join(traceback.format_tb(e.__traceback__))}')
            return None

    def send(self,
             message: Message,
             echo_key_field: Optional[str] = None,
             timeout: Optional[Union[int, float]] = None) -> Optional[ActMessage]:
        """Sends message to conn. Returns sent message's echo if 'echo_key_field' is set, else None.

        Args:
            message (Message): Message to send.
            echo_key_field (str): Key field to catch sent message's echo. If not set, no echo will be returned.
            timeout (int or float, optional): The time (in seconds) during which you are ready to wait for the echo.
                The echo will be returned as soon as it is found, without waiting for the end of the timeout.
                If not set, the gRPC context time will be used.

        Returns:
            Sent message's echo (as ActMessage) or None.
        """
        if not self._act_sender.send_message(message):
            self._grpc_context_manager.cancel_context()
            logger.info('Context was cancelled because of message sending fail')
            return None

        if echo_key_field is not None:
            return self._receive_request_message_echo(message, echo_key_field, timeout)

        return None

    def receive_first_matching(self, *,
                               pass_on: Optional[Union[Callable, Iterable[Callable]]] = None,
                               fail_on: Optional[Union[Callable, Iterable[Callable]]] = None,
                               timeout: Optional[Union[int, float]] = None,
                               check_previous_messages: bool = True) -> ActMessage:
        """Returns the first message (wrapped in ActMessage class) from cache that matches one of the
        filters.

        Args:
            pass_on (lambda or Iterable[lambda]): The filter (lambda or function) through which the messages
                from cache will be filtered out with RequestStatus.SUCCESS.
            fail_on (lambda or Iterable[lambda]): The filter (lambda or function) through which the messages
                from cache will be filtered out with RequestStatus.ERROR.
            timeout (int or float, optional): The time (in seconds) during which you are ready to wait for the message.
                The message will be returned as soon as it is found, without waiting for the end of the timeout.
                If not set, the gRPC context time will be used.
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            ActMessage class instance with received message and status. If no message found,
            ActMessage's 'message' field will be None.
        """

        _, act_message, runtime = self.cache_filter.filter_single_message_within_timeout(
            pass_on=pass_on,
            fail_on=fail_on,
            check_previous_messages=check_previous_messages,
            timeout=timeout
        )

        self._create_and_send_responses_events(receive_method_name='first matching',
                                               runtime=runtime,
                                               act_messages=act_message)
        return act_message

    def receive_first_n_matching(self, *,
                                 n: int,
                                 pass_on: Optional[Union[Callable, Iterable[Callable]]] = None,
                                 fail_on: Optional[Union[Callable, Iterable[Callable]]] = None,
                                 timeout: Optional[Union[int, float]] = None,
                                 check_previous_messages: bool = True) -> List[ActMessage]:
        """Returns a list of first N messages (each wrapped in ActMessage class) from cache that
        match the filters.

        Args:
            pass_on (lambda or Iterable[lambda]): The filter (lambda or function) through which the messages
                from cache will be filtered out with RequestStatus.SUCCESS.
            fail_on (lambda or Iterable[lambda]): The filter (lambda or function) through which the messages
                from cache will be filtered out with RequestStatus.ERROR.
            n (int): A number of filtered messages you want to receive.
            timeout (float, optional): The time (in seconds) during which you are ready to wait for the messages.
                The messages will be returned as soon as N messages are found, without waiting for the end of the
                timeout. If not set, the gRPC context time will be used.
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            A list of ActMessage. If no messages found, empty list will be returned.
        """

        _, act_messages, runtime = self.cache_filter.filter_list_messages_within_timeout(
            pass_on=pass_on,
            fail_on=fail_on,
            timeout=timeout,
            messages_to_receive=n,
            check_previous_messages=check_previous_messages
        )

        self._create_and_send_responses_events(receive_method_name='first n matching',
                                               act_messages=act_messages,
                                               number_of_responses_expected=n,
                                               runtime=runtime)

        return act_messages

    def receive_all_before_matching(self, *,
                                    pass_on: Optional[Union[Callable, Iterable[Callable]]] = None,
                                    fail_on: Optional[Union[Callable, Iterable[Callable]]] = None,
                                    timeout: Optional[Union[int, float]] = None,
                                    check_previous_messages: bool = True) -> List[ActMessage]:
        """Returns a list of messages (each wrapped in ActMessage class) from cache that precede a message
        that matches one of the filters. Message that matching the filter is included in the list.

        Args:
            pass_on (lambda or Iterable[lambda]): The filter (lambda or function) through which the messages
                from cache will be filtered out with RequestStatus.SUCCESS.
            fail_on (lambda or Iterable[lambda]): The filter (lambda or function) through which the messages
                from cache will be filtered out with RequestStatus.ERROR.
            timeout (float, optional): The time (in seconds) during which you are ready to wait for the messages.
                The messages will be returned as soon as the first matching message is found,
                without waiting for the end of the timeout. If not set, the gRPC context time will be used.
            check_previous_messages (bool, optional): Set True if you want to look for the messages throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            A list of ActMessage. If no messages found, empty list will be returned.
        """

        index, act_message, runtime = self.cache_filter.filter_single_message_within_timeout(
            pass_on=pass_on,
            fail_on=fail_on,
            check_previous_messages=check_previous_messages,
            timeout=timeout
        )

        th2_messages = self._cache.get_all_before_matching_from_cache(index=index)
        act_messages = [ActMessage(message=message, status=act_message.status) for message in th2_messages]
        self._create_and_send_responses_events(receive_method_name='all before matching',
                                               act_messages=act_messages,
                                               runtime=runtime)

        return act_messages

    def receive_all_after_matching(self, *,
                                   pass_on: Optional[Union[Callable, Iterable[Callable]]] = None,
                                   fail_on: Optional[Union[Callable, Iterable[Callable]]] = None,
                                   wait_time: Optional[Union[int, float]] = None,
                                   check_previous_messages: bool = True) -> List[ActMessage]:
        """Returns a list of messages (each wrapped in ActMessage class) from cache that follows a message
        that matches one of the filters. Message that matching the filter is included in the list.

        Args:
            pass_on (lambda or Iterable[lambda]): The filter (lambda or function) through which the messages
                from cache will be filtered out with RequestStatus.SUCCESS.
            fail_on (lambda or Iterable[lambda]): The filter (lambda or function) through which the messages
                from cache will be filtered out with RequestStatus.ERROR.
            wait_time (int, optional): The time (in seconds) that the Act will wait before checking the cache for
                the matching message. Defaults to None (cache will be checked immediately and only once).
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            A list of ActMessage. If no messages found, empty list will be returned.
        """

        if wait_time:
            self._grpc_context_manager.wait(wait_time)

        index, act_message, runtime = self.cache_filter.filter_single_message_immediately(
            pass_on=pass_on,
            fail_on=fail_on,
            check_previous_messages=check_previous_messages
        )

        th2_messages = self._cache.get_all_after_matching_from_cache(index=index)
        act_messages = [ActMessage(message=message, status=act_message.status) for message in th2_messages]
        self._create_and_send_responses_events(receive_method_name='all after matching',
                                               act_messages=act_messages,
                                               runtime=runtime)

        return act_messages

    def receive_all_matching(self, *,
                             pass_on: Optional[Union[Callable, Iterable[Callable]]] = None,
                             fail_on: Optional[Union[Callable, Iterable[Callable]]] = None,
                             wait_time: Optional[Union[int, float]] = None,
                             check_previous_messages: bool = True) -> List[ActMessage]:
        """Returns a list of messages (each wrapped in ActMessage class) from cache that matches filters.

        Args:
            pass_on (lambda or Iterable[lambda]): The filter (lambda or function) through which the messages
                from cache will be filtered out with RequestStatus.SUCCESS.
            fail_on (lambda or Iterable[lambda]): The filter (lambda or function) through which the messages
                from cache will be filtered out with RequestStatus.ERROR.
            wait_time (int, optional): The time that the Act will wait before checking the cache for matching messages.
                Defaults to None (cache will be checked immediately).
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            A list of ActMessage. If no messages found, empty list will be returned.
        """

        if wait_time:
            self._grpc_context_manager.wait(wait_time)

        _, act_messages, runtime = self.cache_filter.filter_list_messages_immediately(
            pass_on=pass_on,
            fail_on=fail_on,
            check_previous_messages=check_previous_messages
        )

        self._create_and_send_responses_events(receive_method_name='all matching',
                                               act_messages=act_messages,
                                               runtime=runtime)

        return act_messages

    @property
    def cache_size(self) -> int:
        """Returns a number of messages in cache."""

        return self._cache.size

    def clear_cache(self) -> None:
        """Deletes all messages from cache."""

        self._cache.clear()

    def _receive_request_message_echo(self,
                                      message: Message,
                                      echo_key_field: str,
                                      timeout: Optional[Union[int, float]]) -> Optional[ActMessage]:
        echo_filter = lambda echo: (  # noqa: E731, ECE001
                echo.fields[echo_key_field] == message.fields[echo_key_field]
                and echo.metadata.message_type == message.metadata.message_type
                and echo.metadata.id.direction == Direction.SECOND
                and echo.metadata.id.connection_id.session_alias == message.metadata.id.connection_id.session_alias
        )

        _, act_message, _ = self.cache_filter.filter_single_message_within_timeout(pass_on=echo_filter,
                                                                                   check_previous_messages=False,
                                                                                   timeout=timeout)

        return act_message

    def _create_and_send_responses_root_event(self,
                                              receive_method_name: str,
                                              messages_received: int = 0,
                                              messages_expected: Optional[int] = None) -> EventID:
        if messages_expected in (1, None):
            return self._act_sender.create_and_send_responses_root_event(status=RequestStatus.SUCCESS,
                                                                         event_name=f'Received {messages_received} '
                                                                                    f'message(s) - '
                                                                                    f'{receive_method_name}')
        elif messages_received == messages_expected:
            return self._act_sender.create_and_send_responses_root_event(status=RequestStatus.SUCCESS,
                                                                         event_name=f'Received {messages_received}/'
                                                                                    f'{messages_expected} message(s)'
                                                                                    f' - {receive_method_name}')
        else:
            return self._act_sender.create_and_send_responses_root_event(status=RequestStatus.ERROR,
                                                                         event_name=f'Received {messages_received}/'
                                                                                    f'{messages_expected} message(s)'
                                                                                    f' - {receive_method_name}')

    def _create_and_send_responses_events(self,
                                          receive_method_name: str,
                                          runtime: float,
                                          act_messages: Union[ActMessage, List[ActMessage]],
                                          number_of_responses_expected: Optional[int] = None) -> None:
        responses_root_event = self._create_and_send_responses_root_event(
            receive_method_name=receive_method_name,
            messages_received=ActMessageUtil.get_length(act_messages),
            messages_expected=number_of_responses_expected
        )

        self._act_sender.send_responses_events(
            responses_events=ActMessageUtil.create_events(act_messages=act_messages,
                                                          runtime=runtime,
                                                          responses_root_event=responses_root_event)
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['th2_act', 'th2_act.util']

package_data = \
{'': ['*']}

install_requires = \
['th2-common>=3.9.1,<4.0.0', 'th2-grpc-check1==3.6.0.dev2433840021']

setup_kwargs = {
    'name': 'th2-act',
    'version': '1.0.0.dev3037507623',
    'description': 'Python library with useful tools for creating custom Act implementations',
    'long_description': '# TH2 Act Component (Python)\n\n## Overview\nThis repository is a library for custom gRPC act projects (check [th2-act-template-py](https://github.com/th2-net/th2-act-template-py) \nfor the implementation template). Check the [Wiki](https://github.com/th2-net/th2-act-py/wiki) for instructions and examples.\n\nThe th2-act contains helper methods to create custom \nActHadlers that implements custom gRPC API. Multiple ActHandlers allowed as well.\n\n## Installation\nTo install `th2-act` package run the following command in the terminal:\n```\npip install th2-act\n```\n',
    'author': 'TH2-devs',
    'author_email': 'th2-devs@exactprosystems.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/th2-net/th2-act-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from dataclasses import dataclass
import logging
from typing import List, Optional, Union

from google.protobuf.text_format import MessageToString
import th2_act.util.act_events as events
from th2_grpc_common.common_pb2 import Event, EventID, Message, RequestStatus

logger = logging.getLogger()


@dataclass
class ActMessage:
    """ActMessage is returned by RequestProcessor's 'search' methods. The class instance can contain message
    (if any message received, None otherwise) and status (RequestStatus.SUCCESS or RequestStatus.ERROR).
    """

    message: Optional[Message] = None
    status: Union[str, int] = RequestStatus.ERROR


class ActMessageUtil:

    @staticmethod
    def get_length(act_messages: Union[ActMessage, List[ActMessage]]) -> int:
        if isinstance(act_messages, list):
            return len(act_messages)

        elif act_messages.message is not None:
            return 1

        return 0

    @staticmethod
    def create_events(act_messages: Union[ActMessage, List[ActMessage]],
                      runtime: float,
                      responses_root_event: EventID) -> List[Event]:
        if isinstance(act_messages, list):
            if len(act_messages) > 0:
                return ActMessageUtil._create_succeed_event_list(act_messages=act_messages,
                                                                 runtime=runtime,
                                                                 responses_root_event=responses_root_event)
            else:
                return ActMessageUtil._create_failed_event_list(runtime=runtime,
                                                                responses_root_event=responses_root_event)
        elif isinstance(act_messages, ActMessage):
            if act_messages.message is not None:
                return ActMessageUtil._create_succeed_event(act_message=act_messages,
                                                            responses_root_event=responses_root_event,
                                                            runtime=runtime)
            else:
                return ActMessageUtil._create_failed_event(responses_root_event=responses_root_event,
                                                           runtime=runtime)

        return []

    @staticmethod
    def _create_succeed_event_list(act_messages: List[ActMessage],
                                   runtime: float,
                                   responses_root_event: EventID) -> List[Event]:
        logger.debug('Filtered response messages (in %d s): %s'
                     % (round(runtime, 1),
                        [
                            MessageToString(act_response.message, as_one_line=True)  # type: ignore
                            for act_response in act_messages
                        ])
                     )
        return [
            events.create_receive_response_succeed_event(response=act_message.message,  # type: ignore
                                                         status=act_message.status,
                                                         responses_root_event=responses_root_event)
            for act_message in act_messages
        ]

    @staticmethod
    def _create_failed_event_list(runtime: float,
                                  responses_root_event: EventID) -> List[Event]:
        error_text = f'Messages matching the condition were not found in {round(runtime, 1)} s'
        logger.error(error_text)

        return [
            events.create_receive_response_failed_event(text=error_text,
                                                        responses_root_event=responses_root_event)
        ]

    @staticmethod
    def _create_succeed_event(act_message: ActMessage,
                              responses_root_event: EventID, runtime: float) -> List[Event]:
        logger.debug('Filtered response message (in %d s): %s'
                     % (round(runtime, 1), MessageToString(act_message.message, as_one_line=True)))  # type: ignore

        return [
            events.create_receive_response_succeed_event(response=act_message.message,  # type: ignore
                                                         status=act_message.status,
                                                         responses_root_event=responses_root_event)
        ]

    @staticmethod
    def _create_failed_event(responses_root_event: EventID, runtime: float) -> List[Event]:
        error_text = f'Message matching the condition was not found in {round(runtime, 1)} s'
        logger.error(error_text)

        return [
            events.create_receive_response_failed_event(text=error_text,
                                                        responses_root_event=responses_root_event)
        ]
